//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://emule.sf.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "updownclient.h"
#include "partfile.h"
#include "opcodes.h"
#include "packets.h"
#include "emule.h"

//	members of CUpDownClient
//	which are mainly used for downloading functions 

void CUpDownClient::DrawStatusBar(CDC* dc, RECT* rect, bool onlygreyrect){
	RECT gaprect;
	gaprect.top = rect->top + 2;
	gaprect.bottom = rect->bottom-2;
	gaprect.left = rect->left;
	gaprect.right = rect->right;
	dc->FillRect(&gaprect,&CBrush(RGB(220,220,220)));
	if (onlygreyrect)
		return;
	float blockpixel = (float)(rect->right - rect->left)/((float)(reqfile->GetFileSize()+1)/1024);
	for (int i = 0;i != partcount;i++){
		if (partstatus[i]){
			gaprect.right = rect->left + (uint32)(((float)PARTSIZE*i/1024)*blockpixel);
			if (PARTSIZE*(i+1) > reqfile->GetFileSize())
				gaprect.left = rect->left +  (uint32)((float)((float)reqfile->GetFileSize()/1024)*blockpixel);
			else
				gaprect.left = rect->left +  (uint32)((float)((float)PARTSIZE*(i+1)/1024)*blockpixel);
			if (reqfile->IsComplete(PARTSIZE*i,PARTSIZE*(i+1)-1))
				dc->FillRect(&gaprect,&CBrush(RGB(0,150,0)));
			else
				dc->FillRect(&gaprect,&CBrush(RGB(0,0,0)));
		}
	}
}

bool CUpDownClient::Compare(CUpDownClient* tocomp){
	uint32 test1 = this->GetUserID();
	uint32 test2 = tocomp->GetUserID();
	if (HasLowID())
		return ((this->GetUserID() == tocomp->GetUserID()) && (GetServerIP() == tocomp->GetServerIP()));
	else
		return (this->GetUserID() == tocomp->GetUserID());
}

void CUpDownClient::AskForDownload(){
	lastaskedtime = ::GetTickCount();
	SetDownloadState(DS_CONNECTING);
	TryToConnect();
	SendHelloPacket();
}

void CUpDownClient::SendFileRequest(){
	Packet* packet = new Packet(OP_FILEREQUEST,16);
	memcpy(packet->pBuffer,reqfile->GetFileHash(),16);
	socket->SendPacket(packet,false);
	packet->opcode = OP_SETREQFILEID;
	socket->SendPacket(packet,true);
}

void CUpDownClient::ProcessFileInfo(char* packet,uint32 size){
	CMemFile* data = new CMemFile((BYTE*)packet,size);
	uchar cfilehash[16];
	data->Read(cfilehash,16);
	uint16 namelength;
	data->Read(&namelength,2);
	if (clientfilename)
		delete[] clientfilename;
	clientfilename = new char[namelength+1];
	data->Read(clientfilename,namelength);
	delete data;
	if (memcmp(cfilehash,reqfile->GetFileHash(),16))
		throw "wrong fileid sent";
}

void CUpDownClient::ProcessFileStatus(char* packet,uint32 size){
	CMemFile* data = new CMemFile((BYTE*)packet,size);
	uchar cfilehash[16];
	data->Read(cfilehash,16);
	if (memcmp(cfilehash,reqfile->GetFileHash(),16))
		throw "wrong fileid sent";	
	data->Read(&partcount,2);
	if (partstatus)
		delete[] partstatus;
	bool partsneeded = false;
	if (!partcount){
		partcount = reqfile->GetPartCount();
		partstatus = new uint8[partcount];
		memset(partstatus,1,partcount);
		partsneeded = true;
	}
	else{
		if (reqfile->GetPartCount() != partcount)
			throw "wrong part number";
		partstatus = new uint8[partcount];
		uint16 done = 0;
		while (done != partcount){
			uint8 toread;
			data->Read(&toread,1);
			for (sint32 i = 0;i != 8;i++){
				partstatus[done] = ((toread>>i)&1)? 1:0; 	
				if (partstatus[done] && !reqfile->IsComplete(done*PARTSIZE,((done+1)*PARTSIZE)-1))
					partsneeded = true;
				done++;
				if (done == partcount)
					break;
			}
		}

	}
	theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
	
	if (!partsneeded)
		SetDownloadState(DS_NONEEDEDPARTS);
	else{
		SetDownloadState(DS_ONQUEUE);
		// send downloadrequest
	}
		
	reqfile->NewSrcPartsInfo();
	delete data;
}

bool CUpDownClient::AddRequestForAnotherFile(CPartFile* file){
	for (POSITION pos = otherrequests_list.GetHeadPosition();pos != 0;otherrequests_list.GetNext(pos)){
		if (otherrequests_list.GetAt(pos) == file)
			return false;
	}
	otherrequests_list.AddTail(file);
	return true;
}

void CUpDownClient::SetDownloadState(uint8 news){
	if (substate != news){
		substate = news;
		theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
	}
}