//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://emule.sf.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// ListenSocket.cpp : implementation file
//

#include "stdafx.h"
#include "emule.h"
#include "ListenSocket.h"
#include "opcodes.h"
#include "KnownFile.h"
#include "sharedfilelist.h"
#include "uploadqueue.h"
#include "updownclient.h"


// CClientReqSocket
CClientReqSocket::CClientReqSocket(CPreferences* in_prefs,bool in_connected, CUpDownClient* in_client){
	app_prefs = in_prefs;
	sizereceived = 0;
	sizetoget = 0;
	rbuffer = 0;
	headercomplete = false;
	pendingpacket = 0;
	pendingbuffer = 0;
	isbusy = false;
	client = in_client;
	connected = in_connected;
	disconnected = false;
	theApp.listensocket->AddSocket(this);
	ResetTimeOutTimer();
}


CClientReqSocket::~CClientReqSocket(){
	theApp.listensocket->RemoveSocket(this);
	safe_delete(pendingpacket);
	if (pendingbuffer)
		delete[] pendingbuffer;
	if (rbuffer)
		delete[] rbuffer;
	Close();
}

void CClientReqSocket::ResetTimeOutTimer(){
	timeout_timer = ::GetTickCount();
};

bool CClientReqSocket::CheckTimeOut(){
	if (::GetTickCount() - timeout_timer > CONNECTION_TIMEOUT){
		timeout_timer = ::GetTickCount();
		Disconnect();
		return true;
	}
	return false;
};

void CClientReqSocket::OnClose(int nErrorCode){
	CAsyncSocket::OnClose(nErrorCode);
	Disconnect();
};

void CClientReqSocket::Disconnect(){
	connected = false;
	disconnected = true;
	Close();
	if (!client)
		delete this;
	else
		client->Disconnected();
};

void CClientReqSocket::OnReceive(int nErrorCode){
	ResetTimeOutTimer();
	if  (!sizetoget || !headercomplete){
		if (!sizetoget){
			rbuffer = new char[5];
			headercomplete = false;
			uint32 read = this->Receive(rbuffer,5);
			if (read == SOCKET_ERROR)
				return;
			sizereceived +=read; 
		}
		else
			sizereceived += this->Receive(rbuffer+sizereceived,sizetoget-sizereceived);

		if (sizereceived < 5){
			theApp.emuledlg->AddLogLine(false,"Error in Server:OnReceive! size < 5");
			sizetoget = 5;
			return;
		}
		else{
			Header_Struct* header = (Header_Struct*) rbuffer;
			if (header->eDonkeyID != OP_EDONKEYHEADER){
				theApp.emuledlg->AddLogLine(false,"Debug: Error in Server:OnReceive! header invalid");
				// errorcorrection: try to find next header somewhere		
				for (int i = 0;i != 5; i++){
					if (rbuffer[i] == OP_EDONKEYHEADER){
						memmove(rbuffer,rbuffer+i,5-i);
						sizetoget = 5;
						sizereceived = 5-i;
						return;
					}
				}
				delete[] rbuffer;
				rbuffer = 0;
				sizereceived = 0;
				return;
			}
			else {
				headercomplete = true;
				sizetoget = header->packetlength;
				sizereceived = 0;
				delete[] rbuffer;
				rbuffer = 0;
				return;
			}
		}
	}
	else {
		if (!rbuffer || !sizereceived)
			rbuffer = new char[sizetoget];
		uint32 read  = Receive(rbuffer+sizereceived,sizetoget-sizereceived);
		if (read == SOCKET_ERROR)
			return;
		sizereceived += read;
		if (sizetoget == sizereceived){
			if (!ProcessPacket(rbuffer+1, sizetoget-1, rbuffer[0]))
				return;
			delete[] rbuffer;
			rbuffer = 0;
			sizetoget = 0;
			sizereceived = 0;
		}
	}
}

bool CClientReqSocket::ProcessPacket(char* packet, uint32 size, uint8 opcode){
	try{
		if (!client && opcode != OP_HELLO)
			throw "Asks for something without saying hello";
		switch(opcode){
			case OP_HELLOANSWER:{
				client->ProcessHelloAnswer(packet,size);
				client->ConnectionEstablished();
				break;
			}
			case OP_HELLO:{
				if (!client){
					// create new client to save standart informations
					client = new CUpDownClient(this);
				}
				client->ProcessHelloPacket(packet,size);
				// now we check if we now this client already. if yes this socket will
				// be attached to the known client, the new client will be deleted
				// and the var. "client" will point to the known client.
				// if not we keep our new-constructed client ;)
				if (theApp.uploadqueue->AttachToAlreadyKnown(&client,this) || theApp.downloadqueue->AttachToAlreadyKnown(&client,this)){
					// update the old client informations
					client->ProcessHelloPacket(packet,size);
				}
				// send a response packet with standart informations
				client->SendHelloAnswer();
				client->ConnectionEstablished();

				break;
			}
			case OP_FILEREQUEST:{
				if (size == 16){
					if (!client->GetWaitStartTime())
						client->SetWaitStartTime();
					client->SetUploadState(US_PENDING);
					CKnownFile* reqfile = theApp.sharedfiles->GetFileByID((uchar*)packet);
					if (reqfile){
					// send filename etc
						memcpy(client->reqfileid,packet,16);
						CMemFile* data = new CMemFile();
						data->Write(reqfile->GetFileHash(),16);
						uint16 namelength = strlen(reqfile->GetFileName());
						data->Write(&namelength,2);
						data->Write(reqfile->GetFileName(),namelength);
						Packet* packet = new Packet(data);
						packet->opcode = OP_FILEREQANSWER;
						delete data;
						SendPacket(packet,true);
					//send filestatus
						data = new CMemFile();
						data->Write(reqfile->GetFileHash(),16);
						if (reqfile->IsPartFile())
							((CPartFile*)reqfile)->WritePartStatus(data);
						else{
							uint32 null = 0;
							data->Write(&null,3);
						}
						packet = new Packet(data);
						packet->opcode = OP_FILESTATUS;
						delete data;
						SendPacket(packet,true);
						break;
					}
				}
				throw "invalid packet size";
				break;
			}
			case OP_FILEREQANSWER:{
				client->ProcessFileInfo(packet,size);
				break;
			}
			case OP_FILESTATUS:{
				client->ProcessFileStatus(packet,size);
				break;
			}
			case OP_STARTUPLOADREQ:{
					theApp.uploadqueue->AddClientToQueue(client);
				break;
			}
			case OP_REQUESTPARTS:{
				CMemFile* data = new CMemFile((BYTE*)packet,size);
				uchar reqfilehash[16];
				data->Read(reqfilehash,16);
				Requested_Block_Struct* reqblock1 = new Requested_Block_Struct;
				Requested_Block_Struct* reqblock2 = new Requested_Block_Struct;
				Requested_Block_Struct* reqblock3 = new Requested_Block_Struct;
				data->Read(&reqblock1->StartOffset,4);
				data->Read(&reqblock2->StartOffset,4);
				data->Read(&reqblock3->StartOffset,4);
				data->Read(&reqblock1->EndOffset,4);
				data->Read(&reqblock2->EndOffset,4);
				data->Read(&reqblock3->EndOffset,4);
				memcpy(&reqblock1->FileID,reqfilehash,16);
				memcpy(&reqblock2->FileID,reqfilehash,16);
				memcpy(&reqblock3->FileID,reqfilehash,16);
				if (reqblock1->EndOffset-reqblock1->StartOffset == 0)			
					delete reqblock1;
				else
					client->AddReqBlock(reqblock1);

				if (reqblock2->EndOffset-reqblock2->StartOffset == 0)			
					delete reqblock2;
				else
					client->AddReqBlock(reqblock2);

				if (reqblock3->EndOffset-reqblock3->StartOffset == 0)			
					delete reqblock3;
				else
					client->AddReqBlock(reqblock3);
				delete data;
				break;
			}
			case OP_CANCELTRANSFER:{
				theApp.uploadqueue->RemoveFromUploadQueue(client);
				break;
			}
			//default:
			  
		}
	}
	catch(char* error){
		if (client)
			theApp.emuledlg->AddLogLine(false,
			"Client '%s' (IP:%s) caused an error or did something bad: %s. Disconnecting client!"
			,client->GetUserName(),client->GetFullIP(),error);
		else
			theApp.emuledlg->AddLogLine(false,"A client caused an error or did something bad: %s. Disconnecting client!",error);
		client->SetDownloadState(DS_ERROR);		
		Disconnect();
		return false;
	}
	return true;
}

bool CClientReqSocket::SendPacket(Packet* packet,bool delpacket){
	if (!IsConnected()){
		safe_delete(pendingpacket);
		if (!delpacket){
			Packet* copy = new Packet(packet->opcode,packet->size);
			memcpy(copy->pBuffer,packet->pBuffer,packet->size);
			pendingpacket = copy;
		}
		else
			pendingpacket = packet;
		return false;
	}
	if (Send(packet->GetHeader(),6))
		return false;
	if (packet->pBuffer){
		if (Send(packet->pBuffer,packet->size))
			return false;
	}
	if (delpacket)
		delete packet;
	return true;
}

void CClientReqSocket::OnConnect(int nErrorCode){
	CAsyncSocket::OnConnect(nErrorCode);
	ResetTimeOutTimer();
	if (nErrorCode){
		if (client)
			client->ConnectionFailed();
	}
	else{
		connected = true;
		if (pendingpacket){
			this->SendPacket(pendingpacket,true);
			pendingpacket = 0;
		}
	}
}

void CClientReqSocket::OnSend(int nErrorCode){
	if (pendingbuffer){
		if (!Send(pendingbuffer,pendingblen,0,false)){
			delete[] pendingbuffer;
			pendingbuffer = 0;
		}
		else
			return;
	}
	CAsyncSocket::OnSend(nErrorCode);
	isbusy = false;
}

int CClientReqSocket::Send(const void*lpBuf,int nBufLen,int nFlags,bool safesend){
	ResetTimeOutTimer();
	if (CAsyncSocket::Send(lpBuf,nBufLen,nFlags) == SOCKET_ERROR){
		if (GetLastError() == WSAEWOULDBLOCK){
			isbusy = true;
			if (safesend){
				if (!pendingbuffer){
					pendingbuffer = new char[nBufLen];
					memcpy(pendingbuffer,lpBuf,nBufLen);
					pendingblen = nBufLen;
				}
				else{
					char* buffer = new char[pendingblen+nBufLen];
					memcpy(buffer,pendingbuffer,pendingblen);
					memcpy(&buffer[pendingblen],lpBuf,nBufLen);
					pendingblen += nBufLen;
					delete[] pendingbuffer;
					pendingbuffer = buffer;
				}
			}
			return -1;
		}
		else{
			return -1;
		}
	}
	return 0;
	
}
// CListenSocket
// CListenSocket member functions
CListenSocket::CListenSocket(CPreferences* in_prefs){
	app_prefs = in_prefs;
}

CListenSocket::~CListenSocket(){
	Close();
	KillAllSockets();
}

bool CListenSocket::StartListening(){
	return (this->Create(app_prefs->GetPort()) && this->Listen());
}

void CListenSocket::OnAccept(int nErrorCode){
	//TODO: Block banned IPs here
	CClientReqSocket* newclient = new CClientReqSocket(app_prefs,true);
	this->Accept(*newclient);
}

//ok that sucks pretty much, but I see no better to imlement a timeout for CClientReqSocket
void CListenSocket::Process(){
	POSITION pos2;
	for(POSITION pos1 = socket_list.GetHeadPosition(); ( pos2 = pos1 ) != NULL; ){
       socket_list.GetNext(pos1);
       socket_list.GetAt( pos2 )->CheckTimeOut();
   }
}

void CListenSocket::AddSocket(CClientReqSocket* toadd){
	socket_list.AddTail(toadd);
}

void CListenSocket::RemoveSocket(CClientReqSocket* todel){
	for (POSITION pos = socket_list.GetHeadPosition();pos != 0;socket_list.GetNext(pos))
		if (socket_list.GetAt(pos) == todel){
			socket_list.RemoveAt(pos);
			return;
		}
}

void CListenSocket::KillAllSockets(){
	for (POSITION pos = socket_list.GetHeadPosition();pos != 0;pos = socket_list.GetHeadPosition()){
		CClientReqSocket* cur_socket = socket_list.GetAt(pos);
			if (cur_socket->client)
				delete cur_socket->client;
			else
				delete cur_socket;
	}
}

bool CListenSocket::TooManySockets(){
	if (GetOpenSockets() > app_prefs->GetMaxConnections()){
		return true;
	}
	else
		return false;
}

